# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.Version import __version__

from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='Furious-GUI',
    version=__version__,
    license='GPL v3.0',
    description='A PySide6-based cross platform GUI client that launches your beloved GFW to outer space.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Loren Eteval',
    author_email='loren.eteval@proton.me',
    url='https://github.com/LorenEteval/Furious',
    packages=find_packages(),
    package_data={'Furious': ['Data/**']},
    include_package_data=True,
    install_requires=[
        'PySide6',
        'qdarkstyle',
        'ujson',
        'pybase64',
        'pyqrcode',
        'pypng',
        'ping3',
        'sysproxy; platform_system == "Windows"',
        'darkdetect; platform_system != "Darwin"',
        'darkdetect[macos-listener]; platform_system == "Darwin"',
    ],
    extras_require={
        'go1.20': [
            'Xray-core < 1.8.5',
            'hysteria > 1.3.5',
            'hysteria2 == 2.0.0.1',
            'tun2socks > 2.5.1; platform_system != "Linux"',
        ],
        'go1.21': [
            'Xray-core >= 1.8.5, < 1.8.8',
            'hysteria2 >= 2.0.4',
            'tun2socks > 2.5.1; platform_system != "Linux"',
        ],
        'go1.22': [
            'Xray-core >= 1.8.8',
            'hysteria2 >= 2.0.4',
        ],
    },
    entry_points={
        'gui_scripts': [
            'Furious = Furious.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Topic :: Internet',
        'Topic :: Internet :: Proxy Servers',
    ],
)



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from PySide6.QtWidgets import QApplication

__all__ = [
    'Translatable',
    'SupportConnectedCallback',
    'SupportThemeChangedCallback',
    'SupportExitCleanup',
    'SupportImplicitReference',
    'FastItemDeletionSearch',
]

import logging

logger = logging.getLogger(__name__)


class Translatable:
    ObjectsPool = list()

    def __init__(self, *args, **kwargs):
        self.translatable = kwargs.pop('translatable', True)

        super().__init__(*args, **kwargs)

        Translatable.ObjectsPool.append(self)

    def retranslate(self):
        raise NotImplementedError

    @staticmethod
    def retranslateAll():
        for ob in Translatable.ObjectsPool:
            assert isinstance(ob, Translatable)

            if ob.translatable:
                ob.retranslate()


class SupportConnectedCallback:
    ObjectsPool = list()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        SupportConnectedCallback.ObjectsPool.append(self)

    def disconnectedCallback(self):
        raise NotImplementedError

    def connectedCallback(self):
        raise NotImplementedError

    @staticmethod
    def callConnectedCallback():
        for ob in SupportConnectedCallback.ObjectsPool:
            assert isinstance(ob, SupportConnectedCallback)

            ob.connectedCallback()

    @staticmethod
    def callDisconnectedCallback():
        for ob in SupportConnectedCallback.ObjectsPool:
            assert isinstance(ob, SupportConnectedCallback)

            ob.disconnectedCallback()


class SupportThemeChangedCallback:
    ObjectsPool = list()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        SupportThemeChangedCallback.ObjectsPool.append(self)

    def themeChangedCallback(self, theme: str):
        raise NotImplementedError

    @staticmethod
    def callThemeChangedCallbackUnchecked(theme: str):
        for ob in SupportThemeChangedCallback.ObjectsPool:
            assert isinstance(ob, SupportThemeChangedCallback)

            ob.themeChangedCallback(theme)

    @staticmethod
    def callThemeChangedCallback(theme: str):
        try:
            app = QApplication.instance()

            if app is not None and app.isDarkModeEnabled():
                # Ignore application dark detect system
                logger.info(f'ignore system theme \'{theme}\' changes in dark mode')

                return
        except Exception:
            # Any non-exit exceptions

            pass

        logger.info(f'system theme changed to \'{theme}\'')

        SupportThemeChangedCallback.callThemeChangedCallbackUnchecked(theme)


class SupportExitCleanup:
    ObjectsPool = list()
    VisitedType = dict()

    def __init__(self, *args, **kwargs):
        self.uniqueCleanup = kwargs.pop('uniqueCleanup', True)

        super().__init__(*args, **kwargs)

        SupportExitCleanup.ObjectsPool.append(self)

    def cleanup(self):
        raise NotImplementedError

    @staticmethod
    def cleanupAll():
        for ob in SupportExitCleanup.ObjectsPool:
            assert isinstance(ob, SupportExitCleanup)

            if ob.uniqueCleanup:
                obtype = str(type(ob))

                if not SupportExitCleanup.VisitedType.get(obtype, False):
                    ob.cleanup()

                    SupportExitCleanup.VisitedType[obtype] = True
                else:
                    pass
            else:
                ob.cleanup()


class SupportImplicitReference:
    ObjectsPool = list()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        SupportImplicitReference.ObjectsPool.append(self)


class FastItemDeletionSearch:
    DeletedItem = list()
    DeletedId = dict()

    @staticmethod
    def moveToTrash(item):
        FastItemDeletionSearch.DeletedItem.append(item)
        FastItemDeletionSearch.DeletedId[id(item)] = True

    @staticmethod
    def isInTrash(item) -> bool:
        return id(item) in FastItemDeletionSearch.DeletedId



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any, Sequence

__all__ = ['ItemUpdateProtocol']


class ItemUpdateProtocol:
    def __init__(self, sequence: Sequence, currentIndex: int, currentItem: Any):
        self.sequence = sequence
        self.currentIndex = currentIndex
        self.currentItem = currentItem

    def currentItemDeleted(self, *args, **kwargs) -> bool:
        return self.currentIndex < 0 or self.currentIndex >= len(self.sequence)

    def updateImpl(self, *args, **kwargs):
        raise NotImplementedError

    def updateResult(self):
        if self.currentItemDeleted():
            # Deleted. Do nothing
            return

        if id(self.sequence[self.currentIndex]) == id(self.currentItem):
            self.updateImpl()
        else:
            # Linear find and update
            for index, item in enumerate(self.sequence):
                if id(item) == id(self.currentItem):
                    # Found. Update index
                    self.currentIndex = index
                    self.updateImpl()

                    # Updated
                    break

            # Not found. Do nothing



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility.Constants import *
from Furious.Utility.AppSettings import AppSettings
from Furious.Utility.Utility import runExternalCommand

import os
import sys
import ctypes
import functools
import subprocess

__all__ = [
    'getPythonVersion',
    'getUbuntuRelease',
    'isAdministrator',
    'isVPNMode',
    'isScriptMode',
    'isPythonw',
    'isWindows7',
]


def getPythonVersion():
    return '.'.join(str(info) for info in sys.version_info)


@functools.lru_cache(None)
def getUbuntuRelease() -> str:
    try:
        result = runExternalCommand(
            ['cat', '/etc/lsb-release'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )

        values = dict(
            list(line.split('='))
            for line in filter(lambda x: x != '', result.stdout.decode().split('\n'))
        )

        if values['DISTRIB_ID'] == 'Ubuntu':
            return values['DISTRIB_RELEASE']
        else:
            return ''
    except Exception:
        # Any non-exit exceptions

        return ''


@functools.lru_cache(None)
def isAdministrator() -> bool:
    if PLATFORM == 'Windows':
        return ctypes.windll.shell32.IsUserAnAdmin() == 1
    else:
        return os.geteuid() == 0


def isVPNMode() -> bool:
    return isAdministrator() and AppSettings.isStateON_('VPNMode')


def isScriptMode() -> bool:
    return sys.argv[0].endswith('.py')


def isPythonw() -> bool:
    def isRealFile(file):
        if not hasattr(file, 'fileno'):
            return False

        try:
            tmp = os.dup(file.fileno())
        except Exception:
            # Any non-exit exceptions

            return False
        else:
            os.close(tmp)

            return True

    # pythonw.exe. Also applies to packed GUI application on Windows
    return not isRealFile(sys.__stdout__) or not isRealFile(sys.__stderr__)


def isWindows7() -> bool:
    return PLATFORM == 'Windows' and PLATFORM_RELEASE == '7'



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.Library import *
from Furious.Utility import *

from typing import Union

import time
import logging
import multiprocessing

logger = logging.getLogger(__name__)

__all__ = ['Hysteria1']


def startHysteria1(jsonString, rule, mmdb, msgQueue: multiprocessing.Queue):
    try:
        import hysteria
    except ImportError:
        # Fake running process
        while True:
            time.sleep(1)
    else:
        if hysteria.__version__ <= '1.3.5':
            redirect = False
        else:
            redirect = True

        StdoutRedirectHelper.launch(
            msgQueue, lambda: hysteria.startFromJSON(jsonString, rule, mmdb), redirect
        )


class Hysteria1(CoreProcess):
    class ExitCode:
        ConfigurationError = 23
        RemoteNetworkError = 3
        # Windows shutting down
        SystemShuttingDown = 0x40010004

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def rule(rulePath):
        if isinstance(rulePath, str) and rulePath == '':
            return ''

        try:
            path = getAbsolutePath(str(rulePath))
        except Exception:
            # Any non-exit exceptions

            logger.error('invalid hysteria1 rule path. Fall back to empty')

            return ''

        try:
            with open(path, 'rb') as file:
                data = file.read()

            logger.info(f'hysteria1 rule \'{path}\' load success')

            return data
        except Exception as ex:
            # Any non-exit exceptions

            logger.error(
                f'hysteria1 rule \'{path}\' load failed. {ex}. Fall back to empty'
            )

            return ''

    @staticmethod
    def mmdb(mmdbPath):
        if isinstance(mmdbPath, str) and mmdbPath == '':
            return ''

        try:
            path = getAbsolutePath(str(mmdbPath))
        except Exception:
            # Any non-exit exceptions

            logger.error('invalid hysteria1 mmdb path. Fall back to empty')

            return ''

        try:
            with open(path, 'rb') as file:
                data = file.read()

            logger.info(f'hysteria1 mmdb \'{path}\' load success')

            return data
        except Exception as ex:
            # Any non-exit exceptions

            logger.error(
                f'hysteria1 mmdb \'{path}\' load failed. {ex}. Fall back to empty'
            )

            return ''

    @staticmethod
    def name() -> str:
        return 'Hysteria1'

    @staticmethod
    def version() -> str:
        try:
            import hysteria

            return hysteria.__version__
        except Exception:
            # Any non-exit exceptions

            return '0.0.0'

    def startFromArgs(self, jsonString: str, rule, mmdb, **kwargs) -> bool:
        self.registerCurrentJSONConfig(jsonString)

        return super().start(
            target=startHysteria1,
            args=(jsonString, rule, mmdb, self.msgQueue),
            **kwargs,
        )

    def start(self, config: Union[str, dict], rule, mmdb, **kwargs) -> bool:
        if isinstance(config, str):
            return self.startFromArgs(config, rule, mmdb, **kwargs)
        elif isinstance(config, ConfigurationFactory):
            return self.startFromArgs(config.toJSONString(), rule, mmdb, **kwargs)
        elif isinstance(config, dict):
            try:
                jsonString = UJSONEncoder.encode(config)
            except Exception:
                # Any non-exit exceptions

                return False
            else:
                return self.startFromArgs(jsonString, rule, mmdb, **kwargs)
        else:
            return False



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.QtFramework import *
from Furious.QtFramework import gettext as _

from PySide6 import QtCore
from PySide6.QtWidgets import *

import functools

__all__ = ['LogViewerWindow']

needTrans = functools.partial(needTransFn, source=__name__)

needTrans('Unable to save log')


class SaveErrorMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.saveError = ''

    def customText(self):
        if self.saveError:
            return _('Unable to save log') + f'\n\n{self.saveError}'
        else:
            return _('Unable to save log')

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(self.customText())

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans(
    'Save File',
    'Text files (*.txt);;All files (*)',
    'Error saving log',
)


def saveAsFile(content: str):
    filename, selectedFilter = QFileDialog.getSaveFileName(
        None, _('Save File'), filter=_('Text files (*.txt);;All files (*)')
    )

    if filename:
        try:
            with open(filename, 'w', encoding='utf-8') as file:
                file.write(content)
        except Exception as ex:
            # Any non-exit exceptions

            mbox = SaveErrorMBox(icon=AppQMessageBox.Icon.Critical)
            mbox.saveError = str(ex)
            mbox.setWindowTitle(_('Error saving log'))
            mbox.setText(mbox.customText())

            # Show the MessageBox and wait for user to close it
            mbox.exec()


needTrans(
    'Log Viewer',
    'Save As...',
    'Exit',
    'File',
    'Copy',
    'Select All',
    'Edit',
    'Zoom In',
    'Zoom Out',
    'View',
)


class LogViewerWindow(AppQMainWindow):
    def __init__(self, *args, **kwargs):
        tabTitle = kwargs.pop('tabTitle', '')
        fontFamily = kwargs.pop('fontFamily', '')
        pointSizeSettingsName = kwargs.pop('pointSizeSettingsName', '')

        super().__init__(*args, **kwargs)

        self.setWindowTitle(_('Log Viewer'))

        self.textBrowser = DraculaTextBrowser(
            fontFamily=fontFamily,
            pointSizeSettingsName=pointSizeSettingsName,
        )
        self.textBrowser.setLineWrapMode(DraculaTextBrowser.LineWrapMode.NoWrap)

        self.tabWidget = AppQTabWidget()
        self.tabWidget.addTab(self.textBrowser, tabTitle)

        self.setCentralWidget(self.tabWidget)

        self._fileMenu = AppQMenu(
            AppQAction(
                _('Save As...'),
                callback=lambda: saveAsFile(self.textBrowser.toPlainText()),
            ),
            AppQSeperator(),
            AppQAction(
                _('Exit'),
                callback=lambda: self.hide(),
            ),
            title=_('File'),
            parent=self,
        )

        self._editMenu = AppQMenu(
            AppQAction(
                _('Copy'),
                icon=bootstrapIcon('files.svg'),
                callback=lambda: self.textBrowser.copy(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_C,
                ),
            ),
            AppQSeperator(),
            AppQAction(
                _('Select All'),
                callback=lambda: self.textBrowser.selectAll(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_A,
                ),
            ),
            title=_('Edit'),
            parent=self,
        )

        self._viewMenu = AppQMenu(
            AppQAction(
                _('Zoom In'),
                callback=lambda: self.textBrowser.zoomIn(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_Plus,
                ),
            ),
            AppQAction(
                _('Zoom Out'),
                callback=lambda: self.textBrowser.zoomOut(),
                shortcut=QtCore.QKeyCombination(
                    QtCore.Qt.KeyboardModifier.ControlModifier,
                    QtCore.Qt.Key.Key_Minus,
                ),
            ),
            title=_('View'),
            parent=self,
        )

        self.menuBar().addMenu(self._fileMenu)
        self.menuBar().addMenu(self._editMenu)
        self.menuBar().addMenu(self._viewMenu)

    def appendLine(self, line: str):
        self.textBrowser.appendLine(line)

    def clear(self):
        self.textBrowser.clear()
